#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-23
# @Filename: scripts.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from . import lvmnps_command_parser


if TYPE_CHECKING:
    from src.lvmnps.actor.actor import NPSCommand


__all__ = ["scripts"]


def validate_nps(command: NPSCommand):
    """Checks that the NPS implements scripting."""

    nps = command.actor.nps

    if nps.implementations.get("scripting", False) is False:
        command.fail("Scripting not allowed for this NPS.")
        return False

    return True


@lvmnps_command_parser.group()
def scripts():
    """Handles user scripts."""

    pass


@scripts.command()
@click.argument("SCRIPT", type=str, nargs=-1)
async def run(command: NPSCommand, script: tuple[str, ...]):
    """Runs a user script.

    The first argument is expected to be the name of the script. Additional
    argument will be considered arguments to pass to the script function.

    """

    if not validate_nps(command):
        return

    if len(script) == 0:
        return command.fail("Not enough parameters.")

    nps = command.actor.nps

    script_name = script[0]
    script_args = script[1:]

    try:
        thread_id = await nps.run_script(script_name, *script_args)
    except Exception as err:
        return command.fail(f"Failed executing script {script_name!r}: {err}")

    return command.finish(
        script={
            "name": script_name,
            "args": list(script_args),
            "running": True,
            "thread_id": thread_id,
        }
    )


@scripts.command()
@click.argument("THREAD", type=int, required=False)
async def stop(command: NPSCommand, thread: int | None):
    """Runs a user script.

    If the thread number is not provided, stops all running scripts.

    """

    if not validate_nps(command):
        return

    nps = command.actor.nps

    try:
        await nps.stop_script(thread_num=thread)
    except Exception:
        return command.fail("Failed stopping scripts.")

    return command.finish()


@scripts.command(name="list")
async def list_scripts(command: NPSCommand):
    """Lists running scripts."""

    if not validate_nps(command):
        return

    nps = command.actor.nps

    threads = await nps.list_running_scripts()

    scripts = [
        {"name": threads[id_], "args": [], "running": True, "thread_id": id_}
        for id_ in threads
    ]

    return command.finish(scripts=scripts)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-22
# @Filename: core.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import abc
import asyncio

from typing import Any, Sequence, TypedDict

import httpx
from pydantic import BaseModel, ConfigDict

from lvmnps import log
from lvmnps.exceptions import VerificationError
from lvmnps.tools import get_outlet_by_id, get_outlet_by_name, normalise_outlet_name


__all__ = ["NPSClient", "OutletModel", "OutletArgType"]


class OutletModel(BaseModel):
    """A model for an outlet status."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: int
    name: str
    normalised_name: str = ""
    state: bool = False

    _client: NPSClient | None = None

    def model_post_init(self, __context: Any) -> None:
        self.normalised_name = normalise_outlet_name(self.name)

        return super().model_post_init(__context)

    def set_client(self, nps: NPSClient):
        """Sets the NPS client."""

        self._client = nps

    async def on(self):
        """Sets the state of the outlet to "on"."""

        if not self._client:
            raise RuntimeError("NPS client not set.")

        await self._client.set_state(self, on=True)

    async def off(self):
        """Sets the state of the outlet to "off"."""

        if not self._client:
            raise RuntimeError("NPS client not set.")

        await self._client.set_state(self, on=False)


OutletArgType = OutletModel | int | str | Sequence[str | int | OutletModel]


class ImplementationsDict(TypedDict):
    """Dictionary of NPS implementations."""

    scripting: bool


class NPSClient(abc.ABC):
    """Base NPS client."""

    nps_type: str
    implementations: ImplementationsDict = {"scripting": False}

    def __init__(self):
        self.outlets: dict[str, OutletModel] = {}

        # Time after switching an outlet on during which switching outlets on is
        # delayed to prevent simultaneous inrush currents on power-on time.
        self.delay: float = 1

    async def setup(self):
        """Sets up the power supply, setting any required configuration options."""

        pass

    async def stop(self):
        """Performs any necessary operations to gracefully disconnect from the NPS."""

        pass

    @abc.abstractmethod
    async def verify(self):
        """Checks that the NPS is connected and responding."""

        pass

    @abc.abstractmethod
    async def refresh(self):
        """Refreshes the list of outlets."""

        pass

    def _validate_response(self, response: httpx.Response, expected_code: int = 200):
        """Validates an HTTP response."""

        if response.status_code != expected_code:
            raise VerificationError(
                f"Request returned response with status code {response.status_code}."
            )

    def get(self, outlet: int | str):
        """Retrieves an outlet by ID or name."""

        if isinstance(outlet, int):
            return get_outlet_by_id(self.outlets, outlet)
        elif isinstance(outlet, str):
            return get_outlet_by_name(self.outlets, outlet)
        else:
            raise TypeError("Invalid outlet type. Only int and str are allowed.")

    async def set_state(
        self,
        outlets: OutletArgType,
        on: bool = False,
        off_after: float | None = None,
    ) -> list[OutletModel]:
        """Sets the state of an outlet or list of outlets.

        Parameters
        ----------
        outlets
            An outlet or list of outlets whose state will be set. An outlet
            can be specified by its name, number, or model instance. If a list
            of outlet is provided the behaviour will depend on the client
            implementation. Outlets may be switched concurrently or sequentially,
            with a delay to avoid in-rush currents.
        on
            Whether to turn the outlet on (if ``True``) or off.
        off_after
            Turns off the outlet after the specified number of seconds. Only
            relevant if ``on=True``.

        """

        _outlets: list[OutletModel] = []

        if isinstance(outlets, str) or not isinstance(outlets, Sequence):
            outlets = [outlets]

        for outlet in outlets:
            if isinstance(outlet, str):
                _outlets.append(get_outlet_by_name(self.outlets, outlet))
            elif isinstance(outlet, int):
                _outlets.append(get_outlet_by_id(self.outlets, outlet))
            else:
                _outlets.append(outlet)

        names = [outlet.name for outlet in _outlets]
        log.debug(f"Setting outlets {names} to state on={on}.")
        await self._set_state_internal(_outlets, on=on, off_after=off_after)

        await self.refresh()

        # Gets the updated outlets we modified.
        switched_outlets: list[OutletModel] = []
        for _outlet in _outlets:
            switched_outlets.append(
                get_outlet_by_name(
                    self.outlets,
                    _outlet.normalised_name,
                )
            )

        return switched_outlets

    @abc.abstractmethod
    async def _set_state_internal(
        self,
        outlets: list[OutletModel],
        on: bool = False,
        off_after: float | None = None,
    ):
        """Internal method for setting the outlet state.

        This method is intended to be overridden by each specific implementation.
        All implementations should handle switching single outlets or multiple ones,
        and do it in a way that is safe and efficient given the hardware specifications.

        """

    async def cycle(
        self,
        outlets: OutletArgType,
        delay: float = 3,
    ) -> list[OutletModel]:
        """Turns off the selected outlets and turns them on again after a delay.

        Parameters
        ----------
        outlets
            An outlet or list of outlets whose state will be set. An outlet
            can be specified by its name, number, or model instance.
        delay
            Number of seconds the code will wait before turning on the
            outlets again.

        """

        await self.set_state(outlets, on=False)
        await asyncio.sleep(delay)
        switched_outlets = await self.set_state(outlets, on=True)

        return switched_outlets

    async def run_script(self, name: str, *args, **kwargs) -> int:
        """Runs a user script."""

        raise NotImplementedError()

    async def stop_script(self, thread_num: int | None = None):
        """Stops a running script."""

        raise NotImplementedError()

    async def list_running_scripts(self) -> dict[int, str]:
        """Returns a list of running scripts."""

        raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-22
# @Filename: exceptions.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations


class NPSException(Exception):
    """Base NPS exception."""


class VerificationError(NPSException):
    """Failed to connect to the power supply."""


class ResponseError(NPSException):
    """Invalid response from the power supply API."""


class NPSWarning(UserWarning):
    """Base NPS warning."""



import asyncio
import functools
import logging
import os

import click
from click_default_group import DefaultGroup

from sdsstools.daemonizer import DaemonGroup

from lvmnps.actor.actor import NPSActor


def cli_coro(f):
    """Decorator function that allows defining coroutines with click."""

    if hasattr(asyncio, "coroutine"):
        f = getattr(asyncio, "coroutine")(f)

    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))

    return functools.update_wrapper(wrapper, f)


@click.group(cls=DefaultGroup, default="actor", default_if_no_args=True)
@click.option(
    "-c",
    "--config",
    "config_file",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    help="Path to the user configuration file.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Debug mode.",
)
@click.pass_context
def lvmnps(ctx: click.Context, config_file: str, verbose: bool = False):
    """Network Power Supply actor."""

    ctx.obj = {"verbose": verbose, "config_file": config_file}


@lvmnps.group(cls=DaemonGroup, prog="nps_actor", workdir=os.getcwd())
@click.pass_context
@cli_coro
async def actor(ctx):
    """Runs the actor."""

    config_file = ctx.obj["config_file"]

    lvmnps = NPSActor.from_config(config_file, verbose=ctx.obj["verbose"])

    if lvmnps.log.fh:
        lvmnps.log.fh.setLevel(logging.NOTSET)

    if ctx.obj["verbose"]:
        lvmnps.log.sh.setLevel(logging.NOTSET)
    else:
        lvmnps.log.sh.setLevel(logging.INFO)

    await lvmnps.start()
    await lvmnps.run_forever()


def main():
    lvmnps(auto_envvar_prefix="LVMNPS")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Mingyeong YANG (mingyeong@khu.ac.kr)
# @Date: 2021-03-22
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from sdsstools import get_config, get_logger, get_package_version


NAME = "sdss-lvmnps"


log = get_logger(NAME, use_rich_handler=True)

__version__ = get_package_version(path=__file__, package_name=NAME)


from .actor import NPSActor, NPSCommand
from .nps import DLIClient, NPSClient



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Mingyeong YANG (mingyeong@khu.ac.kr), Florian Briegel (briegel@mpia.de)
# @Date: 2021-03-22
# @Filename: lvmnps/actor/actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib
from os import PathLike

from typing import TYPE_CHECKING

from clu import Command
from clu.actor import AMQPActor
from sdsstools.configuration import Configuration

from lvmnps import __version__
from lvmnps import log as nps_log
from lvmnps.actor.commands import lvmnps_command_parser
from lvmnps.nps.core import NPSClient
from lvmnps.nps.implementations import VALID_NPS_TYPES


if TYPE_CHECKING:
    from sdsstools.logger import SDSSLogger


__all__ = ["NPSActor"]


AnyPath = str | PathLike[str]


def get_nps_from_config(config: Configuration) -> NPSClient:
    """Returns an `.NPSClient` instance from the configuration parameters."""

    if "nps" not in config:
        raise ValueError("nps section does not exist in the configuration.")

    nps_type = config["nps.type"]
    if nps_type is None:
        raise ValueError("nps.type not defined.")

    if nps_type not in VALID_NPS_TYPES:
        raise ValueError(f"Invalid NPS {nps_type}. Valid types are {VALID_NPS_TYPES}.")

    init_parameters = config.get("nps.init_parameters", {})

    if nps_type == "dli":
        from lvmnps.nps.implementations.dli import DLIClient

        return DLIClient(**init_parameters)

    elif nps_type == "netio":
        from lvmnps.nps.implementations.netio import NetIOClient

        return NetIOClient(**init_parameters)

    else:  # pragma: no cover - This should unreachable.
        raise ValueError(f"Invalid NPS {nps_type}. Valid types are {VALID_NPS_TYPES}.")


class NPSActor(AMQPActor):
    """LVM network power switches base actor."""

    parser = lvmnps_command_parser

    def __init__(
        self,
        *args,
        schema: AnyPath | None = None,
        log: SDSSLogger | None = None,
        **kwargs,
    ):
        cwd = pathlib.Path(__file__).parent

        schema = schema or cwd / "schema.json"
        log = log or nps_log

        kwargs["version"] = __version__

        super().__init__(*args, log=log, schema=schema, **kwargs)

        if not isinstance(self.config, Configuration):
            self.config = Configuration(self.config)

        self.nps = get_nps_from_config(self.config)

    async def start(self, **kwargs):  # pragma: no cover
        """Starts the actor."""

        await self.nps.setup()

        return await super().start(**kwargs)

    async def stop(self):
        "Stops the actor."

        await self.nps.stop()

        return await super().stop()


NPSCommand = Command[NPSActor]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Mingyeong YANG (mingyeong@khu.ac.kr)
# @Date: 2021-03-22
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from .actor import NPSActor, NPSCommand



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-23
# @Filename: refresh.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from . import lvmnps_command_parser


if TYPE_CHECKING:
    from src.lvmnps.actor.actor import NPSCommand


__all__ = ["refresh"]


@lvmnps_command_parser.command()
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Does not output status after refreshing",
)
async def refresh(command: NPSCommand, quiet: bool = False):
    """Refreshes the internal list of outlets."""

    nps = command.actor.nps

    await nps.refresh()

    if not quiet:
        await command.child_command("status")

    command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-23
# @Filename: onoff.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from . import lvmnps_command_parser


if TYPE_CHECKING:
    from src.lvmnps.actor.actor import NPSCommand


__all__ = ["on", "off", "cycle"]


class OutletNameParamType(click.ParamType):
    name = "outlet_name"

    def convert(self, value, param, ctx):
        if isinstance(value, int):
            return value

        try:
            return int(value, 10)
        except ValueError:
            return value


@lvmnps_command_parser.command()
@click.argument("OUTLETS", type=OutletNameParamType(), nargs=-1)
@click.option("--off-after", type=float, help="Turns the outlet off after N seconds.")
async def on(
    command: NPSCommand,
    outlets: tuple[str | int, ...],
    off_after: float | None = None,
):
    """Turns an outlet or outlets on.

    OUTLETS can be a single outlet name or outlet ID, or a list of them that will
    be switched on as quickly as possible. An argument that can be casted into
    an integer will be considered an outlet ID.

    """

    nps = command.actor.nps

    outlet_data = await nps.set_state(outlets, on=True, off_after=off_after)

    command.finish(outlets=[outlet.model_dump() for outlet in outlet_data])


@lvmnps_command_parser.command()
@click.argument("OUTLETS", type=OutletNameParamType(), nargs=-1)
async def off(command: NPSCommand, outlets: tuple[str | int, ...]):
    """Turns an outlet or outlets off.

    OUTLETS can be a single outlet name or outlet ID, or a list of them that will
    be switched on as quickly as possible. An argument that can be casted into
    an integer will be considered an outlet ID.

    """

    nps = command.actor.nps

    outlet_data = await nps.set_state(outlets, on=False)

    command.finish(outlets=[outlet.model_dump() for outlet in outlet_data])


@lvmnps_command_parser.command()
@click.argument("OUTLETS", type=OutletNameParamType(), nargs=-1)
@click.option(
    "--delay",
    type=float,
    default=3,
    help="Delay between off an on in seconds",
)
async def cycle(command: NPSCommand, outlets: tuple[str | int, ...], delay: float = 3):
    """Cycles an outlet or outlets.

    OUTLETS can be a single outlet name or outlet ID, or a list of them that will
    be switched on as quickly as possible. An argument that can be casted into
    an integer will be considered an outlet ID.

    """

    nps = command.actor.nps

    outlet_data = await nps.cycle(outlets, delay=delay)

    command.finish(outlets=[outlet.model_dump() for outlet in outlet_data])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-23
# @Filename: netio.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import warnings

from typing import Any

import httpx
from pydantic import ConfigDict, SecretStr
from pydantic.dataclasses import dataclass

from lvmnps import log
from lvmnps.exceptions import NPSWarning, VerificationError
from lvmnps.nps.core import NPSClient, OutletModel
from lvmnps.tools import APIClient


class NetIOOutLetModel(OutletModel):
    """Outlet model for NetIO switches."""

    pass


@dataclass(config=ConfigDict(extra="forbid"))
class NetIOClient(NPSClient):
    """An NPS client for a NetIO switch.

    This implementation uses the JSON API, see
    https://www.netio-products.com/files/NETIO-M2M-API-Protocol-JSON.pdf

    """

    host: str
    port: int = 80
    user: str = "admin"
    password: SecretStr = SecretStr("admin")

    def __post_init__(self):
        super().__init__()

        self.base_url = f"http://{self.host}:{self.port}"
        self.api_client = APIClient(
            self.base_url,
            self.user,
            self.password,
            auth_method="basic",
        )

        self.outlets: dict[str, NetIOOutLetModel] = {}

        self.nps_type = "netio"
        self.implementations = {"scripting": False}

    async def setup(self):
        """Sets up the power supply, setting any required configuration options."""

        log.info("Setting up NetIO switch.")

        try:
            await self.verify()
        except VerificationError as err:
            warnings.warn(
                "Cannot setup NetIO. Power switch "
                f"verification failed with error: {err}",
                NPSWarning,
            )
            return

        await self.refresh()

        log.info("Set up complete.")

    async def verify(self):
        """Checks that the NPS is connected and responding."""

        async with self.api_client as client:
            try:
                response = await client.get(url="/netio.json")
            except httpx.ConnectError as err:
                raise VerificationError(f"Failed to connect to NetIO: {err}")

        self._validate_response(response, 200)

    async def refresh(self):
        """Refreshes the list of outlets."""

        log.debug("Refreshing list of outlets.")

        async with self.api_client as client:
            response = await client.get(url="/netio.json")

        self._validate_response(response)

        data = response.json()
        outlet_json = data["Outputs"]

        log.debug(f"Found {len(outlet_json)} outlets.")

        self.outlets = {}

        for data in outlet_json:
            outlet = NetIOOutLetModel(
                id=data["ID"],
                name=data["Name"],
                state=data["State"],
            )
            outlet.set_client(self)
            self.outlets[outlet.normalised_name] = outlet

    async def _set_state_internal(
        self,
        outlets: list[NetIOOutLetModel],
        on: bool = False,
        off_after: float | None = None,
    ):
        """Sets the state of a list of outlets."""

        outputs: list[dict[str, Any]] = []

        for outlet in outlets:
            outlet_action: dict[str, Any] = {"ID": outlet.id, "Action": int(on)}
            if on is True and off_after is not None:
                outlet_action["Action"] = 3
                outlet_action["Delay"] = off_after * 1000
            outputs.append(outlet_action)

        async with self.api_client as client:
            response = await client.post(
                url="/netio.json",
                json={"Outputs": outputs},
            )

        self._validate_response(response)

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-22
# @Filename: status.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

from . import lvmnps_command_parser


if TYPE_CHECKING:
    from src.lvmnps.actor.actor import NPSCommand


__all__ = ["status"]


@lvmnps_command_parser.command()
async def status(command: NPSCommand):
    """Outputs the status of the network power switch."""

    nps = command.actor.nps

    command.info(nps_type=nps.nps_type)
    command.info(outlet_names=list(nps.outlets))

    outlets = [outlet.model_dump() for outlet in nps.outlets.values()]
    command.info(outlets=outlets)

    command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-22
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING

import httpx
from pydantic import SecretStr
from pydantic.dataclasses import dataclass
from typing_extensions import Literal

from lvmnps import log


if TYPE_CHECKING:
    from lvmnps.nps.core import OutletModel


__all__ = [
    "APIClient",
    "normalise_outlet_name",
    "get_outlet_by_name",
    "get_outlet_by_id",
]


@dataclass
class APIClient:
    """A wrapper around ``httpx.AsyncClient`` to yield a new client."""

    base_url: str
    user: str
    password: SecretStr

    auth_method: Literal["digest", "basic"] = "digest"

    def __post_init__(self):
        self.client: httpx.AsyncClient | None = None
        self.lock = asyncio.Lock()

    async def __aenter__(self):
        """Yields a new client."""

        await self.lock.acquire()

        log.debug(f"Creating async client to {self.base_url!r} with digest.")

        if self.auth_method == "digest":
            auth = httpx.DigestAuth(self.user, self.password.get_secret_value())
        elif self.auth_method == "basic":
            auth = (self.user, self.password.get_secret_value())
        else:
            raise ValueError(f"Invalud authentication method {self.auth_method!r}.")

        self.client = httpx.AsyncClient(
            auth=auth,
            base_url=self.base_url,
            headers={},
        )

        return self.client

    async def __aexit__(self, exc_type, exc, tb):
        """Closes the client."""

        self.lock.release()

        if self.client and not self.client.is_closed:
            log.debug("Closing async client.")
            await self.client.aclose()


def normalise_outlet_name(name: str):
    """Returns a normalised name for an outlet."""

    return name.lower().replace(" ", "_")


def get_outlet_by_name(outlet_data: dict[str, OutletModel], name: str):
    """Gets an outlet from a list of outlets.

    Parameters
    ----------
    outlet_data
        The mapping of outlet name to outlet model data.
    name
        The name of the outlet to retrieve.

    Returns
    -------
    outlet
        The outlet matching the input name.

    Raises
    ------
    ValueError
        If the outlet cannot be found.

    """

    normalised_name = normalise_outlet_name(name)

    if normalised_name in outlet_data:
        return outlet_data[normalised_name]

    raise ValueError(f"Cannot find outlet with name {name!r}.")


def get_outlet_by_id(outlet_data: dict[str, OutletModel], id: int):
    """Gets an outlet by id.

    Parameters
    ----------
    outlet_data
        The mapping of outlet name to outlet model data.
    id
        The id of the outlet to retrieve.

    Returns
    -------
    outlet
        The outlet matching the id.

    Raises
    ------
    ValueError
        If the outlet cannot be found.

    """

    for outlet in outlet_data.values():
        if outlet.id == id:
            return outlet

    raise ValueError(f"Cannot find outlet with id {id!r}.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-22
# @Filename: dli.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import warnings

import httpx
from pydantic import ConfigDict, SecretStr
from pydantic.dataclasses import dataclass

from lvmnps import log
from lvmnps.exceptions import NPSWarning, ResponseError, VerificationError
from lvmnps.nps.core import NPSClient, OutletModel
from lvmnps.tools import APIClient


__all__ = ["DLIClient", "DLIOutletModel"]


class DLIOutletModel(OutletModel):
    """A model for a DLI outlet status."""

    index: int = 0
    physical_state: bool = False
    transient_state: bool = False
    critical: bool = False
    locked: bool = False
    cycle_delay: float | None = None


@dataclass(config=ConfigDict(extra="forbid"))
class DLIClient(NPSClient):
    """An NPS client for a Digital Loggers switch."""

    host: str
    port: int = 80
    user: str = "admin"
    password: SecretStr = SecretStr("admin")
    api_route: str = "restapi/"

    def __post_init__(self):
        super().__init__()

        self.base_url = f"http://{self.host}:{self.port}/{self.api_route}"
        self.api_client = APIClient(
            self.base_url,
            self.user,
            self.password,
            auth_method="digest",
        )

        self.outlets: dict[str, DLIOutletModel] = {}

        self.nps_type = "dli"
        self.implementations = {"scripting": True}

    async def setup(self):
        """Sets up the power supply, setting any required configuration options."""

        log.info("Setting up DLI switch.")

        try:
            await self.verify()
        except VerificationError as err:
            warnings.warn(
                "Cannot setup DLI. Power switch "
                f"verification failed with error: {err}",
                NPSWarning,
            )
            return

        async with self.api_client as client:
            # Change in-rush delay to 1 second.
            log.debug("Setting sequence delay to 1 second.")
            response = await client.put(
                url="/relay/sequence_delay/",
                data={"value": 1},
                headers={"X-CSRF": "x"},
            )
            self._validate_response(response, 204)

        await self.refresh()

        log.info("Set up complete.")

    async def verify(self):
        """Checks that the NPS is connected and responding."""

        async with self.api_client as client:
            try:
                response = await client.get(url="/", headers={"Range": "dli-depth=1"})
            except httpx.ConnectError as err:
                raise VerificationError(f"Failed to connect to DLI: {err}")

        # 206 because we have asked for the API to only do depth=1
        self._validate_response(response, 206)

    async def refresh(self):
        """Refreshes the list of outlets."""

        log.debug("Refreshing list of outlets.")

        url = "/relay/outlets/"
        async with self.api_client as client:
            response = await client.get(url=url)

        self._validate_response(response)

        data = response.json()
        log.debug(f"Found {len(data)} outlets.")

        self.outlets = {}

        for outlet_id in range(1, len(data) + 1):
            outlet_data = data[outlet_id - 1]
            outlet_data["id"] = outlet_id
            outlet_data["index"] = outlet_id - 1

            outlet = DLIOutletModel(**outlet_data)
            outlet.set_client(self)
            self.outlets[outlet.normalised_name] = outlet

    async def _set_state_internal(
        self,
        outlets: list[DLIOutletModel],
        on: bool = False,
        off_after: float | None = None,
    ):
        """Sets the state of a list of outlets."""

        outlet_indices = [outlet.index for outlet in outlets]

        # Use a matrix URI to set all the states at once.
        outlet_path = "=" + ",".join(map(str, outlet_indices))

        async with self.api_client as client:
            response = await client.put(
                url=f"/relay/outlets/{outlet_path}/state/",
                data={"value": on},
                headers={"X-CSRF": "x"},
            )
            self._validate_response(response, 207)

        if on is True and off_after is not None:
            await asyncio.sleep(off_after)
            return await self._set_state_internal(outlets, on=False)

        return

    async def list_scripts(self):
        "Retrieves the list of user scripts." ""

        async with self.api_client as client:
            response = await client.get(url="/script/user_functions/")
            self._validate_response(response, 200)

        return list(response.json())

    async def run_script(self, name: str, *args, check_exists: bool = True) -> int:
        """Runs a user script.

        Parameters
        ----------
        name
            The script name.
        args
            Arguments with which to call the script function.
        check_exists
            If ``True``, checks that the script exists in the DLI before
            executing it.

        Returns
        -------
        thread_num
            The thread identifier for the running script.

        """

        if check_exists:
            scripts = await self.list_scripts()
            if name not in scripts:
                raise ValueError(f"Unknown user function {name!r}.")

        data = {"user_function": name}
        if len(args) > 0:
            args_comma = ", ".join(map(str, args))
            data["source"] = f"{name}({args_comma})"

        async with self.api_client as client:
            response = await client.post(
                url="/script/start/",
                json=[data],
                headers={"X-CSRF": "x"},
            )

            if response.status_code == 409:
                raise ResponseError("Invalid function name or argument")

            self._validate_response(response, 200)

        thread_num = int(response.json())

        log.info(f"Script {name!r} is running as thread {thread_num}")

        return thread_num

    async def stop_script(self, thread_num: int | None = None):
        """Stops a running script.

        Parameters
        ----------
        thread_num
            The thread to stop. If not specified, stops all threads.

        """

        if thread_num is None:
            log.info("Stopping all script threads.")
        else:
            log.info(f"Stopping script thread {thread_num}.")

        async with self.api_client as client:
            response = await client.post(
                url="/script/stop/",
                json=["all" if thread_num is None else str(thread_num)],
                headers={"X-CSRF": "x"},
            )

            self._validate_response(response, 200)

    async def list_running_scripts(self) -> dict[int, str]:
        """Returns a mapping of running thread number to script name."""

        threads: dict[int, str] = {}

        async with self.api_client as client:
            response = await client.get(url="/script/threads/")
            self._validate_response(response, 200)

        for thread, description in response.json().items():
            script_name = description["label"].split(" ")[0]
            threads[int(thread)] = script_name

        return threads



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-22
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations


__all__ = [
    "DLIClient",
    "DLIOutletModel",
    "NetIOClient",
    "NetIOOutLetModel",
    "VALID_NPS_TYPES",
]


VALID_NPS_TYPES: list[str] = ["dli", "netio"]


from .dli import DLIClient, DLIOutletModel
from .netio import NetIOClient, NetIOOutLetModel
